package org.apache.nutch.util;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.apache.hadoop.mapred.JobConf;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import java.io.File;
import java.io.FileOutputStream;
import java.net.InetAddress;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicIntegerArray;
import java.util.concurrent.ConcurrentHashMap;
public class HostsUtil {
	private ConcurrentHashMap<String,String[]> hostsPermInfo;//hosts permanent info: timeout,underflow,hostcount
	String[]fields=new String[]{"timeout","underflow","hostcount"};
	double minTimeout=10000;
	double maxTimeout=4*minTimeout;
	double inc=1000;
	double dec=500;
	int hostCount=2000;
	String[]defaultValues=new String[]{""+minTimeout,"false",""+hostCount};
	String fileName;
	
	private AtomicInteger count;
	private AtomicInteger currentsegmentnum;
	private AtomicIntegerArray segCounts;
	int maxNumSegments;
    private ConcurrentHashMap<String,int[]> hostCounts;

	public HostsUtil(){
		hostsPermInfo = new ConcurrentHashMap<String,String[]>();

		fileName="/nutch/hosts_conf.xml";
		loadFromFile();		
	}

	public void loadTempVars(int maxNumSegments_1){
		count=new AtomicInteger(0);
		currentsegmentnum=new AtomicInteger(0);
		maxNumSegments=maxNumSegments_1;
		segCounts = new AtomicIntegerArray(new int[maxNumSegments_1]);
		for(int i=0;i<segCounts.length();i++)
			segCounts.set(i, 0);
		hostCounts = new ConcurrentHashMap<String,int[]>();
	}
	int get(AtomicInteger i){
		return i.get();
	}
	public synchronized int getCount(){
		return get(count);
	}
	public synchronized void incCount(){
		count.incrementAndGet();
	}
	public synchronized int getCurrentsegmentnum(){
		return get(currentsegmentnum);
	}
	public synchronized void incCurrentsegmentnum(){
		currentsegmentnum.incrementAndGet();
		if(currentsegmentnum.get()==maxNumSegments)
			currentsegmentnum.set(0);
	}
	public synchronized int getSegCounts(int i){
		return segCounts.get(i);
	}
	public synchronized int[] getHostCount(String hostordomain){
        int[] hostCount = hostCounts.get(hostordomain);
        //host count: {a,b,c,d} means that from this host there are a urls in segment 0 and b urls in seg 1 and ...
        if (hostCount == null) {
            hostCount = new int[maxNumSegments];
            for(int kl=0;kl<hostCount.length;kl++)
                    hostCount[kl]=0;
            hostCounts.put(hostordomain, hostCount);
        }
        return hostCount;
	}
	public synchronized void incSegCount(int i){
		segCounts.incrementAndGet(i);
	}
	void loadFromFile(){
		try {
			File fXmlFile = new File(fileName);
			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.parse(fXmlFile);
			doc.getDocumentElement().normalize();
			
			NodeList nList = doc.getElementsByTagName("host");
			for (int temp = 0; temp < nList.getLength(); temp++) {
				Node nNode = nList.item(temp);
				if (nNode.getNodeType() == Node.ELEMENT_NODE) {
					Element eElement = (Element) nNode;
					NodeList nlList = null;
					
					String host="";
					
					nlList=eElement.getElementsByTagName("name").item(0).getChildNodes();
					host = ((Node)(nlList.item(0))).getNodeValue();
					
					String []obj=new String[fields.length];
					for(int k=0;k<fields.length;k++){
						try{
							nlList=eElement.getElementsByTagName(fields[k]).item(0).getChildNodes();
							try{
								obj[k]=((Node)(nlList.item(0))).getNodeValue();
							}
							catch(Exception ee){
								obj[k]=defaultValues[k];
							}
						}catch(Exception j){}
					}
					hostsPermInfo.put(host, obj);
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	public void saveFile(){
		Set set = hostsPermInfo.entrySet();
		Iterator it = set.iterator();
		java.io.File f=new java.io.File(fileName);
		try{
			if(!f.exists())
				f.createNewFile();
			}catch(Exception ipo){}
		try{
			java.io.FileOutputStream fs=new java.io.FileOutputStream(f,false);
			fs.write("".getBytes());
			fs=new FileOutputStream(f,true);
			fs.write("<hosts>\n".getBytes());

			while(it.hasNext()) {
				Map.Entry entry = (Map.Entry)it.next();
				fs.write(("<host>\n").getBytes());
				
				fs.write(("<name>").getBytes());
				fs.write((entry.getKey().toString()+"").getBytes());
				fs.write(("</name>\n").getBytes());
				
				String[]val=(String[])entry.getValue();
				for(int k=0;k<fields.length;k++){
					fs.write(("<"+fields[k]+">").getBytes());
					fs.write((val[k]).getBytes());
					fs.write(("</"+fields[k]+">").getBytes());
				}
				
				fs.write(("</host>\n").getBytes());
			}
			fs.write("</hosts>\n".getBytes());
			fs.close();
		}
		catch(Exception iuu){}
	}

	String getValue(String key,String field){
		String[]vals=hostsPermInfo.get(key);
		return getVal(vals,field); 
	}
	String getVal(String[]vals,String field){
		if(vals==null)
			return null;
		for(int i=0;i<fields.length;i++){
			if(fields[i].compareTo(field)==0)
				return vals[i];
		}
		return null;
	}
	public int getTimeout(String url){
		//If host is not in hostsTimeout, add <host , minTimeout> to hostsTimeout
		try{
			String host=normHost(new URL(url));
			String t=getValue(host,"timeout");
			if(t==null){
				hostsPermInfo.put(host, defaultValues);
				return (int)minTimeout;
			}
			return Double.valueOf(t).intValue();
		}
		catch(Exception ee){
			return (int)minTimeout;
		}
	}
	public int getMaxHostCount(String url){
		try{
			String host=normHost(new URL(url));
			String t=getValue(host,"hostcount");
			if(t==null){
				hostsPermInfo.put(host, defaultValues);
				return hostCount;
			}
			return Integer.valueOf(t);
		}
		catch(Exception ee){
			return hostCount;
		}
	}

	public boolean getUnderFlow_Url(String url){
		try{
			String host=normHost(new URL(url));
			if(host==null)
				return false;
			return getUnderFlow_Host(host);
		}
		catch(Exception ee){
			return false;
		}
	}
	boolean getUnderFlow_Host(String host){
		String t=getValue(host,"underflow");
		if(t==null){
			hostsPermInfo.put(host, defaultValues);
			return false;
		}
		return Boolean.valueOf(t);
	}
	public synchronized void incrementTimeout(String url){
		try{
			String host=normHost(new URL(url));
			String[]vals=hostsPermInfo.get(host);			
			double timeout=Double.valueOf(getVal(vals,"timeout"));
			timeout=timeout+inc;
			if (timeout > maxTimeout)
				timeout= maxTimeout;
			for(int i=0;i<vals.length;i++){
				if("timeout".compareTo(fields[i])==0)
					vals[i]=timeout+"";
			}
			hostsPermInfo.put(host,vals);
		}
		catch(Exception ee){
		}
	}
	public synchronized void decrementTimeout(String url){
		try{
			String host=normHost(new URL(url));
			String[]vals=hostsPermInfo.get(host);			
			double timeout=Double.valueOf(getVal(vals,"timeout"));
			timeout=timeout-dec;
			if (timeout < minTimeout)
				timeout= minTimeout;
			for(int i=0;i<vals.length;i++){
				if("timeout".compareTo(fields[i])==0)
					vals[i]=timeout+"";
			}
			hostsPermInfo.put(host,vals);
		}
		catch(Exception ee){
		}
	}
	String normHost(URL url){
		String host="";
		try{
			host=url.getHost();
		}
		catch(Exception e){
			return url.toString();
		}
		
		try{
			InetAddress addr = InetAddress.getByName(host);
			String IP=addr.getHostAddress();
			return IP;
		}catch(Exception eer){}
		
		return host;
	}
}

